import importlib
import json
from http import HTTPStatus
from typing import Any, Dict

import typer
import uvicorn
import yaml
from apidaora import (
    ContentType,
    Response,
    RoutedControllerTypeHint,
    appdaora,
    route,
)
from jsondaora import dataclasses

from asyncapi import Spec
from asyncapi.schema import type_as_jsonschema


def main(
    api_module: str = typer.Option('', envvar='ASYNCAPI_MODULE'),
    host: str = typer.Option('0.0.0.0', envvar='ASYNCAPI_HOST'),
    port: int = typer.Option(5000, envvar='ASYNCAPI_PORT'),
) -> None:
    spec = getattr(importlib.import_module(api_module), 'spec')
    start(spec, host, port)


def start(spec: Spec, host: str, port: int) -> None:
    app = appdaora(
        [build_yaml_spec_controller(spec), build_json_spec_controller(spec)]
    )
    uvicorn.run(app, host=host, port=port)


def build_yaml_spec_controller(spec: Spec) -> RoutedControllerTypeHint:
    @route.get('/asyncapi.yaml')
    def controller() -> Response:
        return Response(
            status=HTTPStatus.OK,
            content_type=ContentType.APPLICATION_YAML,
            body=yaml.dump(spec_asdict(spec)),
            headers=(),
        )

    return controller


def build_json_spec_controller(spec: Spec) -> RoutedControllerTypeHint:
    @route.get('/asyncapi.json')
    def controller() -> Response:
        dict_spec = spec_asdict(spec)
        return Response(
            status=HTTPStatus.OK,
            content_type=ContentType.APPLICATION_JSON,
            body=json.dumps(dict_spec, indent=2).encode(),
            headers=(),
        )

    return controller


def spec_asdict(spec: Spec) -> Dict[str, Any]:
    dict_spec: Dict[str, Any] = dataclasses.asdict(spec)
    dict_spec.pop('default_content_type')
    dict_spec['defaultContentType'] = 'application/json'

    for server in dict_spec['servers'].values():
        server.pop('name')

    if dict_spec['components'] is None:
        dict_spec.pop('components')

    for channel_dict, channel in zip(
        dict_spec['channels'].values(), spec.channels.values()
    ):
        channel_dict.pop('name')
        channel_dict['subscribe']['message'].pop('content_type')

        if 'operation_id' in channel_dict['subscribe']:
            channel_dict['subscribe']['operationId'] = channel_dict[
                'subscribe'
            ].pop('operation_id')

        if channel.subscribe.message.payload:
            channel_dict['subscribe']['message'][
                'payload'
            ] = type_as_jsonschema(channel.subscribe.message.payload)

    return dict_spec


def run() -> None:
    typer.run(main)


if __name__ == '__main__':
    run()



class AsyncApiError(Exception):
    ...


class ReferenceNotFoundError(AsyncApiError):
    ...


class InvalidChannelError(AsyncApiError):
    ...


class OperationIdNotFoundError(AsyncApiError):
    ...


class ChannelOperationNotFoundError(AsyncApiError):
    ...


class UrlOrModuleRequiredError(AsyncApiError):
    ...


class InvalidJsonSchemaTypeError(AsyncApiError):
    ...


class InvalidMessageError(AsyncApiError):
    ...



import dataclasses
from enum import Enum
from typing import Any, Callable, Dict, Optional, Type, get_type_hints


@dataclasses.dataclass
class Info:
    title: str
    version: str
    description: str


class ProtocolType(Enum):
    KAFKA = 'kafka'
    REDIS = 'redis'
    POSTGRES = 'postgres'


@dataclasses.dataclass
class Server:
    name: str
    url: str
    protocol: ProtocolType
    description: str


@dataclasses.dataclass
class Message:
    name: str
    title: str
    summary: str
    content_type: str
    payload: Optional[Type[Any]]


@dataclasses.dataclass
class Subscribe:
    message: Message
    operation_id: Optional[str] = None


@dataclasses.dataclass
class Channel:
    name: str
    description: str
    subscribe: Subscribe


@dataclasses.dataclass
class Components:
    messages: Optional[Dict[str, Message]] = None
    schemas: Optional[Dict[str, Any]] = None


@dataclasses.dataclass
class Specification:
    info: Info
    servers: Dict[str, Server]
    channels: Dict[str, Channel]
    components: Optional[Components] = None
    default_content_type: Optional[str] = None


@dataclasses.dataclass(init=False)
class Spec(Specification):
    def __init__(
        self,
        title: str,
        description: str = '',
        version: str = '1.0.0',
        **servers: str,
    ) -> None:
        self.info = Info(title, version, description)
        self.servers = {}
        self.channels = {}

        for server_name, url in servers.items():
            protocol, uri = url.split('://')
            self.servers[server_name] = Server(
                server_name, uri, ProtocolType(protocol), description=''
            )

    def subscribe(
        self,
        *,
        channel_name: str,
        channel_description: str = '',
        message_name: str = '',
        message_title: str = '',
        message_summary: str = '',
    ) -> Callable[..., Callable[..., Any]]:
        def decorator(subscbriber: Callable[..., Any]) -> Callable[..., Any]:
            message_type = get_type_hints(subscbriber).get('message')
            message = Message(
                name=message_name or channel_name,
                title=message_title,
                summary=message_summary,
                content_type='application/json',
                payload=message_type,
            )
            self.channels[channel_name] = subscbriber.__channel__ = Channel(  # type: ignore
                channel_name,
                channel_description,
                Subscribe(message, subscbriber.__name__),
            )
            return subscbriber

        return decorator



import asyncio
import dataclasses
import logging
from typing import Any, Callable, Dict, Tuple

import orjson
from broadcaster import Broadcast
from jsondaora import DeserializationError, asdataclass, dataclass_asjson

from .entities import Specification
from .exceptions import (
    ChannelOperationNotFoundError,
    InvalidChannelError,
    InvalidMessageError,
    OperationIdNotFoundError,
)


OperationsTypeHint = Dict[Tuple[str, str], Callable[..., Any]]


@dataclasses.dataclass
class AsyncApi:
    spec: Specification
    operations: OperationsTypeHint
    broadcast: Broadcast
    republish_error_messages: bool = True
    logger: logging.Logger = logging.getLogger(__name__)

    async def connect(self) -> None:
        await self.broadcast.connect()

    def payload(self, channel_id: str, **message: Any) -> Any:
        type_ = self.payload_type(channel_id)

        if type_:
            return asdataclass(message, type_)

        return message

    async def publish_json(
        self, channel_id: str, message: Dict[str, Any]
    ) -> None:
        await self.broadcast.publish(
            channel=channel_id,
            message=self.parse_message(
                channel_id, self.payload(channel_id, **message)
            ).decode(),
        )

    async def publish(self, channel_id: str, message: Any) -> None:
        await self.broadcast.publish(
            channel=channel_id,
            message=self.parse_message(channel_id, message).decode(),
        )

    async def listen_all(self) -> None:
        tasks = []

        for channel_id in self.spec.channels.keys():
            task = asyncio.create_task(self.listen(channel_id))
            task.add_done_callback(task_callback)
            tasks.append(task)

        await asyncio.gather(*tasks)

    async def listen(self, channel_id: str) -> None:
        try:
            operation_id = self.spec.channels[
                channel_id
            ].subscribe.operation_id
        except KeyError:
            raise InvalidChannelError(channel_id)

        if operation_id is None:
            raise ChannelOperationNotFoundError(channel_id)

        async with self.broadcast.subscribe(channel=channel_id) as subscriber:
            async for event in subscriber:
                try:
                    json_message = orjson.loads(event.message)
                    payload = self.payload(channel_id, **json_message)

                    coro = self.operations[(channel_id, operation_id)](payload)

                    if asyncio.iscoroutine(coro):
                        await coro

                except (orjson.JSONDecodeError, DeserializationError):
                    raise

                except KeyError:
                    raise OperationIdNotFoundError(operation_id)

                except Exception as error:
                    if not self.republish_error_messages:
                        raise

                    self.logger.warning(
                        f'error={type(error).__name__}; '
                        f"message={event.message[:100]}"
                    )
                    await self.publish(channel_id, payload)

    def parse_message(self, channel_id: str, message: Any) -> Any:
        try:
            type_ = self.spec.channels[channel_id].subscribe.message.payload
        except KeyError:
            raise InvalidChannelError(channel_id)

        if type_:
            if not isinstance(message, type_):
                raise InvalidMessageError(message, type_)

            v = dataclass_asjson(message)
            return v

        return message

    def payload_type(self, channel_id: str) -> Any:
        try:
            return self.spec.channels[channel_id].subscribe.message.payload
        except KeyError:
            raise InvalidChannelError(channel_id)


def task_callback(future: Any) -> None:
    future.result()



"""
asyncapi
"""

__version__ = '0.2.0'
import importlib
import io
from collections import deque
from typing import Any, Dict, Optional

import requests
import yaml
from broadcaster import Broadcast
from jsondaora import jsonschema_asdataclass

from .api import AsyncApi, OperationsTypeHint
from .entities import (
    Channel,
    Components,
    Info,
    Message,
    ProtocolType,
    Server,
    Specification,
    Subscribe,
)
from .exceptions import ReferenceNotFoundError


def build_api(
    path: str,
    server: Optional[str] = None,
    module_name: str = '',
    republish_errors: bool = True,
) -> AsyncApi:
    spec = build_spec(load_spec_dict(path))
    return build_api_from_spec(spec, module_name, server, republish_errors)


def build_api_auto_spec(
    module_name: str,
    server: Optional[str] = None,
    republish_errors: bool = True,
) -> AsyncApi:
    spec = getattr(importlib.import_module(module_name), 'spec')
    return build_api_from_spec(spec, module_name, server, republish_errors)


def build_api_from_spec(
    spec: Specification,
    module_name: str,
    server: Optional[str],
    republish_errors: bool,
) -> AsyncApi:
    operations = build_channel_operations(spec, module_name)

    if server is None:
        server = tuple(spec.servers.keys())[-1]

    protocol = spec.servers[server].protocol.value
    url = spec.servers[server].url
    broadcast = Broadcast(f'{protocol}://{url}')
    return AsyncApi(
        spec, operations, broadcast, republish_error_messages=republish_errors
    )


def build_channel_operations(
    spec: Specification, module_name: str
) -> OperationsTypeHint:
    if module_name:
        return {
            (channel_name, channel.subscribe.operation_id): getattr(
                importlib.import_module(module_name),
                channel.subscribe.operation_id,
            )
            for channel_name, channel in spec.channels.items()
            if channel.subscribe.operation_id
        }

    return {}


def load_spec_dict(path: str) -> Dict[str, Any]:
    spec: Dict[str, Any]

    if path.startswith('http'):
        request = requests.get(path)
        request.raise_for_status()

        if path.endswith('.json'):
            spec = request.json()
        else:
            spec = yaml.safe_load(io.BytesIO(request.content))

    else:
        spec = yaml.safe_load(open(path))

    return spec


def build_spec(spec: Dict[str, Any]) -> Specification:
    fill_refs(spec)

    return Specification(
        info=Info(**spec['info']),
        servers={
            server_name: Server(
                name=server_name,
                protocol=ProtocolType(server_spec.pop('protocol')),
                **server_spec,
            )
            for server_name, server_spec in spec['servers'].items()
        },
        channels={
            channel_name: Channel(
                name=channel_name,
                subscribe=Subscribe(
                    message=Message(
                        content_type=channel_spec['subscribe']['message'].get(
                            'contentType', spec.get('defaultContentType')
                        ),
                        payload=jsonschema_asdataclass(
                            channel_spec['subscribe']['message'][
                                'name'
                            ].replace(' ', ''),
                            channel_spec['subscribe']['message'].get(
                                'payload'
                            ),
                        ),
                        **{
                            k: v
                            for k, v in channel_spec['subscribe'][
                                'message'
                            ].items()
                            if k != 'contentType' and k != 'payload'
                        },
                    ),
                    operation_id=channel_spec.pop('subscribe').get(
                        'operationId', None
                    ),
                ),
                **channel_spec,
            )
            for channel_name, channel_spec in spec['channels'].items()
        },
        components=Components(
            messages={
                msg_id: Message(
                    content_type=message.get(
                        'contentType', spec.get('defaultContentType')
                    ),
                    payload=jsonschema_asdataclass(
                        message['name'].replace(' ', ''),
                        message.get('payload'),
                    ),
                    **{
                        k: v
                        for k, v in message.items()
                        if k != 'contentType' and k != 'payload'
                    },
                )
                for msg_id, message in spec['components']['messages'].items()
            }
            if 'messages' in spec['components']
            else None,
            schemas=spec['components'].get('schemas'),
        )
        if 'components' in spec
        else None,
        default_content_type=spec.get('defaultContentType'),
    )


def fill_refs(
    spec: Dict[str, Any], full_spec: Optional[Dict[str, Any]] = None
) -> None:
    if full_spec is None:
        full_spec = spec

    for key, value in spec.items():
        if isinstance(value, dict):
            while '$ref' in value:
                spec[key] = value = dict_from_ref(spec[key]['$ref'], full_spec)

            fill_refs(value, full_spec)


def dict_from_ref(ref: str, spec: Dict[str, Any]) -> Dict[str, Any]:
    ref_keys = deque(ref.strip('#/').split('/'))

    while ref_keys and isinstance(spec, dict):
        spec = spec.get(ref_keys.popleft(), {})

    if not spec:
        raise ReferenceNotFoundError(ref)

    return spec
